//------------------------------------------------------------------------------
// CERN BE-CO-HT
// GN4124 core for PCIe FMC carrier
// http://www.ohwr.org/projects/gn4124-core
//------------------------------------------------------------------------------
//
// unit name:   main
//
// description: This is a simple example testbench, to demonstrate how to use
// the SystemVerilog BFM of the GN4124 to perform simple accesses over wishbone.
//
// The testbench simply connects the wishbone master of the GN4124 to its own
// DMA configuration wishbone slave and attaches a pre-initialised dummy RAM
// with a wishbone interface to the pipelined DMA interface in order to perform
// a DMA read.
//
//------------------------------------------------------------------------------
// Copyright CERN 2019
//------------------------------------------------------------------------------
// Copyright and related rights are licensed under the Solderpad Hardware
// License, Version 2.0 (the "License"); you may not use this file except
// in compliance with the License. You may obtain a copy of the License at
// http://solderpad.org/licenses/SHL-2.0.
// Unless required by applicable law or agreed to in writing, software,
// hardware and materials distributed under this License is distributed on an
// "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
// or implied. See the License for the specific language governing permissions
// and limitations under the License.
//------------------------------------------------------------------------------

`timescale 1ns/1ps

`include "gn4124_bfm.svh"

import wishbone_pkg::*;

module main;
   reg clk_125m = 0;
   reg clk_62m5 = 0;

   logic gn4124_irq;


   string testphase     = "null:= no phase set yet";
   string test_subphase = "ha!";


   t_wishbone_master_in  wb_in, wb_dma_in, wb_mem_in;
   t_wishbone_master_out wb_out, wb_dma_out, wb_mem_out;

   always #4ns clk_125m <= ~clk_125m;
   always #23ns clk_62m5 <= ~clk_62m5;

   logic rst_125m_n;
   logic rst_62m5_n;

   logic rst_gn4124_n;
   logic clk_gn4124;

   logic wb_dma_clk;
   logic wb_dma_rst_n;

   initial begin
      rst_125m_n = 0;
      rst_62m5_n = 0;
      #80ns;
      rst_125m_n = 1;
      rst_62m5_n = 1;
   end

   IGN4124PCIMaster i_gn4124 ();

   xwb_gn4124_core
     DUT (
          .rst_n_a_i          (i_gn4124.rst_n),
          .p2l_clk_p_i        (i_gn4124.p2l_clk_p),
          .p2l_clk_n_i        (i_gn4124.p2l_clk_n),
          .p2l_data_i         (i_gn4124.p2l_data),
          .p2l_dframe_i       (i_gn4124.p2l_dframe),
          .p2l_valid_i        (i_gn4124.p2l_valid),
          .p2l_rdy_o          (i_gn4124.p2l_rdy),
          .p_wr_req_i         (i_gn4124.p_wr_req),
          .p_wr_rdy_o         (i_gn4124.p_wr_rdy),
          .rx_error_o         (i_gn4124.rx_error),
          .vc_rdy_i           (i_gn4124.vc_rdy),
          .l2p_clk_p_o        (i_gn4124.l2p_clk_p),
          .l2p_clk_n_o        (i_gn4124.l2p_clk_n),
          .l2p_data_o         (i_gn4124.l2p_data),
          .l2p_dframe_o       (i_gn4124.l2p_dframe),
          .l2p_valid_o        (i_gn4124.l2p_valid),
          .l2p_edb_o          (i_gn4124.l2p_edb),
          .l2p_rdy_i          (i_gn4124.l2p_rdy),
          .l_wr_rdy_i         (i_gn4124.l_wr_rdy),
          .p_rd_d_rdy_i       (i_gn4124.p_rd_d_rdy),
          .tx_error_i         (i_gn4124.tx_error),
          .dma_irq_o          (dma_irq),
          .irq_p_i            (1'b0),
          .irq_p_o            (gn4124_irq),
          .status_o           (),
          .wb_master_clk_i    (clk_125m),
          .wb_master_rst_n_i  (rst_125m_n),
          .wb_master_i        (wb_in),
          .wb_master_o        (wb_out),
          .wb_dma_cfg_clk_i   (clk_125m),
          .wb_dma_cfg_rst_n_i (rst_125m_n),
          .wb_dma_cfg_i       (wb_out),
          .wb_dma_cfg_o       (wb_in),
          .wb_dma_dat_clk_i   (wb_dma_clk),
          .wb_dma_dat_rst_n_i (wb_dma_rst_n),
          .wb_dma_dat_i       (wb_dma_in),
          .wb_dma_dat_o       (wb_dma_out)
          );

   assign wb_dma_clk = clk_125m;
   assign wb_dma_rst_n = rst_125m_n;

   xwb_dpram #
     (
      .g_size                  (16384),
      .g_slave1_interface_mode (1), // 1 = PIPELINED
      .g_slave2_interface_mode (1),
      .g_slave1_granularity    (1), // 1 = WORD
      .g_slave2_granularity    (1)
      )
   MEM (
        .rst_n_i   (wb_dma_rst_n),
        .clk_sys_i (wb_dma_clk),
        .slave1_i  (wb_dma_out),
        .slave1_o  (wb_dma_in),
        .slave2_i  (wb_mem_out),
        .slave2_o  (wb_mem_in)
        );

   CBusAccessor acc;

   task val_check(string name, uint32_t addr, val, expected);
      if (val != expected)
        begin
           $display();
           $display("Simulation FAILED");
           $fatal(1, "%s error at address 0x%.2x. Expected 0x%.8x, got 0x%.8x",
                  name, addr, expected, val);
        end
   endtask // val_check

   task reg_check(uint32_t addr, expected);
      uint64_t val;
      acc.read(addr, val);
      val_check("Register read-back", addr, val, expected);
   endtask // reg_check

   task mem_check(uint32_t addr, expected);
      uint64_t val;
      i_gn4124.host_mem_read(addr, val);
      val_check("Memory read-back", addr, val, expected);
   endtask // reg_check

   task check_irq_status;
      //  Check irq status
      reg_check('h04, 'h04);
      if (dma_irq != 1'b1)
        $fatal(1, "dma irq should be 1");
   endtask

   task clear_irq;
      acc.write('h04, 'h04);
      reg_check('h04, 'h00);
      if (dma_irq != 1'b0)
        $fatal(1, "dma irq should be 0");
   endtask




   task t_setup_dma( input [63:0] host_addr,
                     input [31:0] local_addr,
                     input [31:0] length,
                    // input [31:0] attributes,
                     input [63:0] next_chained_descr_addr,
                     input bit more_in_chain,
                     input bit w_nrd_sel);
      automatic logic [31:0] attributes;
      attributes = 'h0;
      attributes[1] = more_in_chain;    // doc has these the wrong way round!!!
      attributes[0] = w_nrd_sel;        // doc has these the wrong way round!!!
      acc.write('h08, local_addr[ 0+:32]);                // carrier local addr
      acc.write('h0c, host_addr[ 0+:32]);                 // hstartL
      acc.write('h10, host_addr[32+:32]);                 // hstartH
      acc.write('h14, length);                            // count
      acc.write('h18, next_chained_descr_addr[ 0+:32]);   // nextL
      acc.write('h1C, next_chained_descr_addr[32+:32]);   // nextH
      acc.write('h20, attributes);                              // attrib
   endtask


   task t_start_dma();
      acc.write('h00, 'h01); // start
//      bits 3:2 are byte_ordering
//      bit 1 is "abort"
// bit 0 is tart
   endtask


   task t_add_chaining_descriptor(
                     input [63:0] host_addr,
                     input [31:0] local_addr,
                     input [31:0] length,
                     //input [31:0] attributes,
                     input [63:0] next_chained_descr_addr,
                     input [63:0] start_address_in_host_memory,   // note the verlog wrapper adds 'h2000_0000
                     input        more_in_chain,
                     input        w_nrd_sel);
      logic [31:0] attributes;
      attributes = 'h0;
      attributes[0] = w_nrd_sel;
      attributes[1] = more_in_chain;
      i_gn4124.host_mem_write_noaddr_twiddles(start_address_in_host_memory + 'h0000, local_addr); // remote address
      i_gn4124.host_mem_write_noaddr_twiddles(start_address_in_host_memory + 'h0004, host_addr[ 0+:32]); // hstartL
      i_gn4124.host_mem_write_noaddr_twiddles(start_address_in_host_memory + 'h0008, host_addr[32+:32]); // hstartH
      i_gn4124.host_mem_write_noaddr_twiddles(start_address_in_host_memory + 'h000C, length); // count
      i_gn4124.host_mem_write_noaddr_twiddles(start_address_in_host_memory + 'h0010, next_chained_descr_addr[ 0+:32]); // nextL
      i_gn4124.host_mem_write_noaddr_twiddles(start_address_in_host_memory + 'h0014, next_chained_descr_addr[32+:32]); // nextH
      i_gn4124.host_mem_write_noaddr_twiddles(start_address_in_host_memory + 'h0018, attributes); // attrib
   endtask




   task automatic t_get_device_mem(ref uint32_t wrd_addr, output [31:0] data);
      data[ 0+:8] = MEM.U_DPRAM.gen_splitram.U_RAM_SPLIT.ram0[wrd_addr];
      data[ 8+:8] = MEM.U_DPRAM.gen_splitram.U_RAM_SPLIT.ram1[wrd_addr];
      data[16+:8] = MEM.U_DPRAM.gen_splitram.U_RAM_SPLIT.ram2[wrd_addr];
      data[24+:8] = MEM.U_DPRAM.gen_splitram.U_RAM_SPLIT.ram3[wrd_addr];
   endtask


   task automatic t_blat_device_mem(input uint32_t wrd_addr, input [31:0] data);
     MEM.U_DPRAM.gen_splitram.U_RAM_SPLIT.ram0[wrd_addr] = data[ 0+:8];
     MEM.U_DPRAM.gen_splitram.U_RAM_SPLIT.ram1[wrd_addr] = data[ 8+:8];
     MEM.U_DPRAM.gen_splitram.U_RAM_SPLIT.ram2[wrd_addr] = data[16+:8];
     MEM.U_DPRAM.gen_splitram.U_RAM_SPLIT.ram3[wrd_addr] = data[24+:8];
   endtask



   task automatic t_compare_memories(input [63:0]    pcie_start_addr,   // note needs to be 8B aligned
                                     input [31:0]    wb_start_addr,
                                     input  int      words_to_compare,
                                     output uint32_t bad_comparisons,
                                     output uint32_t good_comparisons,
                                     input  bit      verbose,
                                     input  bit      no_twiddles = 1'b0);
   uint32_t     addr_cnt;
   uint64_t     system_dat;
   logic [31:0] wb_mem_dat;
   logic [31:0] difference;
   bit          error_detected;
   uint32_t     wb_addr;
   uint64_t     pcie_addr;

   bad_comparisons  = 0;
   good_comparisons = 0;

   if (verbose) begin
      $write("Checking Expected memory contents against what is in the wishbone target memory\n");
      $write("________________________________________________________________________________\n");
      $write("| Addr (byte) | wishbone   | mem content |     PCIE_ADDR      |  error |  cnt  |\n");
      $write("|-------------+------------+-------------+--------------------+--------+-------|\n");
   end

   for (addr_cnt = 'h00; addr_cnt < words_to_compare; addr_cnt += 1)begin
      wb_addr   = wb_start_addr/4   + addr_cnt ;
      pcie_addr = pcie_start_addr + (addr_cnt * 4);

      t_get_device_mem(       wb_addr,  wb_mem_dat);

      if (1'b1 == no_twiddles ) begin
         i_gn4124.host_mem_read_noaddr_twiddles(pcie_addr, system_dat);
      end else begin
         i_gn4124.host_mem_read(pcie_addr, system_dat);
      end

      difference = wb_mem_dat ^ system_dat[0+:32];

       if (difference !== 32'h0) begin
         error_detected = 1'b1;
      end else begin
         error_detected = 1'b0;
      end

      if (1'b1 ==  error_detected) begin
         bad_comparisons++;
      end else begin
         good_comparisons++;
      end

      if (verbose) begin
         $write("| 0x:%08x | 0x%08x | 0x%08x  | 0x%016x |   %1b    | %05d |",
                  4* wb_addr, wb_mem_dat,
                                      system_dat[31:0],
                                                  pcie_addr,
                                                             error_detected,
                                                                       addr_cnt
               );
         if (1'b1 ==  error_detected) begin
            $write(" *** mismatch 0x%08x\n", difference);
         end else begin
            $write("\n");
         end
      end
   end

   if (verbose) begin
      $write("------------------------------------------------------------------------------\n ");
   end
   endtask



// The tests themselves are in the following include file
   `include "testcases.svh"


// ************************************************************************************************
// * Main body of the testbench that acutally drives and tests the DUT
// ************************************************************************************************
   initial begin
      automatic int ntest = 1;
      const     int tests = 13;

      @(posedge i_gn4124.ready);

      acc = i_gn4124.get_accessor();

      acc.set_default_xfer_size(4);

      @(posedge clk_125m);


      // Each test is contained in the testcases.svh file, and can be run atomically without needing
      // memory contents or configuration from previous tests.
      // Ideally this would also include configuring the BFM for each test & resets etc but that was a stage too far!
      // The tests run should also depend on if the DMA engine is enabled (generic)

      // basic tests
      t_test_simple_wishbone_rw( ntest, tests);

      // Wishbone DMA tests
      t_test_unchecked_dma_read_aborted(ntest, tests);
      t_test_256B_dma_write(ntest, tests);
      t_test_2x128B_reads(ntest, tests);
      t_test_256B_read(ntest, tests);
      t_test_dma_rd_with_wishbone_stalls(ntest, tests);
      t_test_8KiB_chained_DMA_WRITE(ntest, tests);
      t_test_8KiB_Byteswaps(ntest, tests);
      t_test_32bit_addr_overflow_rd(ntest, tests);


      t_test_mixed_chaining(ntest, tests);    // commented out as it will fail!

      //t_test_32bit_addr_overflow_wr(ntest, tests);   /// TBD should this test pass?:
          //                              it demonstrates a 4KB Addressing boundary crossed!!!!
          //                               and a 32-bit => 64bit addr....
          // This can be masked in Software (due to sensible addresses => DMA),
          //           but it was fixed for the other direction...


      $display();
      $display("Simulation PASSED");

      $finish;
   end // initial begin

endmodule // main