

// -------------------------------------------
//  TESTCASE:    Simple wishbone RW rest.
//      Purpose:
//            Sanity checking that we have access to wihbone CSR functioning...
//      Method:
//         Use 2 of the addr registers in the DMA control register space as a scratchpad to test read and writes
//         Passes if readback matches expected value.
// --------------------------------------------
   task automatic t_test_simple_wishbone_rw(ref   int testno,
                                            input int num_tests
                                           );
      uint32_t addr, val, expected;

      @(posedge clk_125m);

      // ---------------------------------
      $write("Test %0d/%0d: simple read/write accesses over Wishbone: ",
             testno++, num_tests);
      testphase = "simple read/write accesses over Wishbone";

      // Verify simple read/writes over wishbone
      reg_check('h0, 'h0);

      // use registers as scratchpad to test RW accesses...
      //    EXTEND to validate BE functionality..
      acc.write('h0c, 'hffacce55);
      acc.write('h10, 'h1badcafe);

      reg_check('h0c, 'hffacce55);
      reg_check('h10, 'h1badcafe);

      // Reset all DMA config registers
      for (addr = 'h00; addr <= 'h20; addr += 4)
        begin
           acc.write(addr, 'h0);
        end

      reg_check('h0c, 'h0);
      reg_check('h10, 'h0);

      repeat(2) @(posedge clk_125m);

      $write("PASS\n");
   endtask



// -------------------------------------------
//  TESTCASE:    t_TEST_unchecked_dma_read_aborted
//      Purpose:
//            Sanity checking DMA reads and abort functinoality.
//      Method:
//          NOTE: NOT SELF CHECKING!!!!!!!!!!!!!
// --------------------------------------------
   task automatic t_test_unchecked_dma_read_aborted(ref   int ntest,
                                                    input int tests
                                                   );
      testphase = "128B read over DMA, abort after first read: ";
      $write("Test %0d/%0d: %s", ntest++, tests, testphase);

      if (dma_irq != 1'b0)
        $fatal(1, "dma irq should be 0");

      acc.write('h14, 'h80); // count
      acc.write('h00, 'h01); // start

      // wait for transfer to start
      @(posedge i_gn4124.l2p_valid);

      repeat(2) @(posedge clk_125m);

      // Test abort feature
      acc.write('h00, 'h02);
      reg_check('h04, 'h03);
      acc.write('h00, 'h00);

      repeat(2) @(posedge clk_125m);

      $write("PASS\n");

   endtask



// -------------------------------------------
//  TESTCASE:    t_test_256B_dma_write
//      Purpose:
//           BASic self-checking DMA write (Card to sys_mem)
//      Method:
//          NOTE: self cehcking of memory contents...
// --------------------------------------------
   task automatic t_test_256B_dma_write(ref   int ntest,
                                         input int tests
                                         );
      uint32_t addr,expected, from_mem ;
      bit      error_detected;
      uint32_t badcomp, goodcomp;

      logic [31:0] rand_wdat = $urandom();


   //   rand_wdat = 32'h80000020;

      testphase = "test_256B_dma_write: ";
      $write("Test %0d/%0d: %s", ntest++, tests, testphase);


      // Setup data in BFM memory
      for (addr = 'h00; addr < 'h40; addr += 1)
        i_gn4124.host_mem_write(4 * addr, rand_wdat - addr);

       //  Setup & Start the DMA
        t_setup_dma(  .host_addr               ('h2000_0000)
                     ,.local_addr              ('h00)
                     ,.length                  ('h100)
                     ,.more_in_chain           (0)
                     ,.w_nrd_sel               (1)
                     ,.next_chained_descr_addr ('hdeadbeefbeefbeef)
                   );

      t_start_dma();

      @(posedge dma_irq);
      check_irq_status;
      clear_irq;

      repeat(4) @(posedge clk_125m);

      error_detected = 1'b0;
      for (addr = 'h00; addr < 'h40; addr += 1) begin
           expected = rand_wdat - addr;
           t_get_device_mem(addr, from_mem);
           if (from_mem !== expected) begin
              error_detected = 1'b1;
            end
       end

       if (error_detected)
         t_compare_memories('h00000000, 'h0, 64, badcomp, goodcomp, 1);

       if (error_detected != 1'b1) begin
            $write("PASS \n");
       end else begin
            $write("FAIL  -- IRQ was OK But data integrity failed!!!\n");
       end
      #1us;
   endtask








// -------------------------------------------
//  TESTCASE:    t_test_2x128B_reads
//      Purpose:
//           BASic self-checking DMA read (sys_mem to card)
//      Method:
//          NOTE: self checking of memory contents...
// --------------------------------------------
   task automatic t_test_2x128B_reads(ref   int ntest,
                                         input int tests
                                         );
      uint32_t addr,expected, from_mem ;
      bit      error_detected;
      logic [31:0] rand_wdat = $urandom();


      testphase = "test_2x128B reads: ";
      $write("Test %0d/%0d: %s", ntest++, tests, testphase);

     // rand_wdat = 32'h80000020;

      // TBD: make this "even more" randomised data?
      for (addr = 'h00; addr < 'h40; addr += 1)
        t_blat_device_mem(addr, rand_wdat - addr);

      t_add_chaining_descriptor( .host_addr                   ('h2000_0080)
                                ,.local_addr                  ('h80)
                                ,.length                      ('h80)
                                ,.next_chained_descr_addr     ('h0)
                                ,.start_address_in_host_memory('h20020000)
                                ,.more_in_chain               (1'b0)
                                ,.w_nrd_sel                   (1'b0)
                               );

      t_setup_dma( .host_addr               ('h2000_0000)
                  ,.local_addr              ('h00)
                  ,.length                  ('h80)
                  ,.more_in_chain           (1)
                  ,.w_nrd_sel               (0)
                  ,.next_chained_descr_addr ('h20020000)
                 );

      t_start_dma();

      @(posedge dma_irq);

      check_irq_status;
      clear_irq;

      for (addr = 'h00; addr < 'h40; addr += 1)
        begin
           expected = rand_wdat - addr;
           mem_check(4 * addr, expected);         /// TODO: is this the correct type of memory checking!!!!
        end

      repeat(4) @(posedge clk_125m);

      $write("PASS\n");
   endtask




// -------------------------------------------
//  TESTCASE:    t_test_256B_read
//      Purpose:
//           BASic self-checking DMA read (sys_mem to card)
//      Method:
//          NOTE: self checking of memory contents...
// --------------------------------------------
   task automatic t_test_256B_read(ref   int ntest,
                                         input int tests
                                         );
      uint32_t addr,expected, from_mem ;
      bit      error_detected;

      logic [31:0] rand_wdat = $urandom();


      testphase = "test_256B_read : ";
      $write("Test %0d/%0d: %s", ntest++, tests, testphase);

      for (addr = 'h00; addr < 'h40; addr += 1)
        t_blat_device_mem(addr, rand_wdat - addr*7);

        t_setup_dma(  .host_addr               ('h2000_0000)
                     ,.local_addr              ('h00)
                     ,.length                  ('h100)
                    // ,.attributes              ('h03) //.2
                     ,.more_in_chain           (0)
                     ,.w_nrd_sel               (0)
                     ,.next_chained_descr_addr ('hffeff_0123_4567_19a0)
                   );
       t_start_dma();

      @(posedge dma_irq);

      check_irq_status;
      clear_irq;

      for (addr = 'h00; addr < 'h40; addr += 1)
        begin
           expected = rand_wdat - addr*7;
           mem_check(4 * addr, expected);
        end

      repeat(4) @(posedge clk_125m);

      $write("PASS\n");
   endtask




// -------------------------------------------
//  TESTCASE:    t_test_dma_rd_with_wishbone_stalls
//      Purpose:
//            check the core WRT stalls on the wichbone bus
//      SELF_CHECKING:
//            GROSS check - targets are pre-used and data is non-unique
//      Method:
//      Notes:
//
// --------------------------------------------
task automatic t_test_dma_rd_with_wishbone_stalls(ref   int ntest,
                                                  input int tests
                                                 );

      uint32_t addr, val, expected;
      int stall_cycle[14];

      logic [31:0] rand_wdat = $urandom();

      testphase = "256B DMA reads with wishbone stalling: ";
      $write("Test %0d/%0d: %s", ntest++, tests, testphase);


      for (addr = 'h00; addr < 'h40; addr += 1) begin
        t_blat_device_mem(addr, rand_wdat - addr);
      end


      t_setup_dma(  .host_addr               ('h2000_0000)
                   ,.local_addr              ('h00)
                   ,.length                  ('h100)
                   ,.more_in_chain           (0)              // above it is 0 -- is it intentionally using a chain?
                   ,.w_nrd_sel               (0)
                   ,.next_chained_descr_addr ('h20020000)
                 );

      stall_cycle = '{0,1,2,14,15,16,17,30,31,32,33,61,62,63};
      foreach(stall_cycle[i])
        begin
           //acc.write('h00, 'h01); // start

         $sformat(testphase, "256B DMA reads with wishbone stalling:  (stall_cycle = %0d)", stall_cycle[i]);

           t_start_dma();

           @(posedge wb_dma_out.stb);
           repeat(stall_cycle[i]) @(posedge clk_125m);
           force DUT.cmp_wrapped_gn4124.dma_stall_i = 1'b1;
           @(posedge clk_125m);
           force DUT.cmp_wrapped_gn4124.dma_ack_i = 1'b0;
           @(posedge clk_125m);
           release DUT.cmp_wrapped_gn4124.dma_stall_i;
           @(posedge clk_125m);
           release DUT.cmp_wrapped_gn4124.dma_ack_i;

           @(posedge dma_irq);

           check_irq_status;
           clear_irq;

           for (addr = 'h00; addr < 'h40; addr += 1)  begin
                expected = rand_wdat - addr;
                mem_check(4 * addr, expected);
           end

           repeat(4) @(posedge clk_125m);
        end

      $write("PASS\n");
endtask




// -------------------------------------------
//  TESTCASE:    t_test_2x4KiB_chained_DMA_WRITE
//      Purpose:
//            legacy: checks write chains...
//      SELF_CHECKING:
//            NONE - no timeout    does not even wait to confirm transfer finished...
//      Method:
//      Notes:
//
// --------------------------------------------
task automatic t_test_2x4KiB_chained_DMA_WRITE(ref   int ntest,
                                                  input int tests
                                                 );

      uint32_t addr, val, expected;

      // ---------------------------------
      testphase = "2x4KiB chained DMA write: ";
      $write("Test %0d/%0d: %s", ntest++, tests, testphase);

      // Setup data in BFM memory
      for (addr = 'h00; addr < 'h800; addr += 1)
        i_gn4124.host_mem_write(4 * addr, 32'h80000020 - addr);

      t_add_chaining_descriptor( .host_addr                    ('h2000_1000)
                                ,.local_addr                   ('h1000)
                                ,.length                       ('h1000)
                                ,.next_chained_descr_addr      ('h0)
                                ,.start_address_in_host_memory ('h2002_0000)
                                ,.more_in_chain                (1'b0)
                                ,.w_nrd_sel                    (1'b1)
                               );


        t_setup_dma(  .host_addr               ('h2000_0000)
                     ,.local_addr              ('h00)
                     ,.length                  ('h1000)
                     ,.more_in_chain           (0)
                     ,.w_nrd_sel               (1)
                     ,.next_chained_descr_addr ('h2002_0000)
                   );

       t_start_dma();

      @(posedge dma_irq);

      check_irq_status;
      clear_irq;

      // cehck data

      repeat(4) @(posedge clk_125m);

      $write("PASS\n");

      // wait for WB transfer to finish
      #5us;


endtask






 // -------------------------------------------
//  TESTCASE:    t_test_8KiB_chained_DMA_WRITE
//      Purpose:
//            legacy: checks write chains...
//      SELF_CHECKING:
//            NONE - no timeout    does not even wait to confirm transfer finished...
//      Method:
//      Notes:
//
// --------------------------------------------
task automatic t_test_8KiB_chained_DMA_WRITE(ref   int ntest,
                                                  input int tests
                                                 );

      uint32_t addr, val, expected, from_mem;
      bit      error_detected;
      uint32_t badcomp, goodcomp;

      logic [31:0] rand_wdat = $urandom();

      // ---------------------------------
      testphase = "8KiB chained DMA write: ";
      $write("Test %0d/%0d: %s", ntest++, tests, testphase);


       // Setup data in BFM memory
      for (addr = 'h00; addr < 'h800; addr += 1) begin
        i_gn4124.host_mem_write(4 * addr, rand_wdat - addr);
      end


      t_add_chaining_descriptor( .host_addr                   (64'h20001800)
                                ,.local_addr                  (32'h1800)
                                ,.length                      ('h800)
                                ,.next_chained_descr_addr     (32'hDEADBEEf)   // OK I tweaked this but...
                                ,.start_address_in_host_memory('h2002_0000)
                                ,.more_in_chain               (1'b0)
                                ,.w_nrd_sel                   (1'b1)
                               );

        t_setup_dma(  .host_addr               (64'h2000_0000)
                     ,.local_addr              ('h00)
                     ,.length                  ('h1800)
                     ,.more_in_chain           (1)
                     ,.w_nrd_sel               (1)
                     ,.next_chained_descr_addr (64'h20020000)
                   );

       t_start_dma();

      @(posedge dma_irq);

      check_irq_status;
      clear_irq;

      repeat(4) @(posedge clk_125m);

      // do I need to wait longer here???

      // Check data
      error_detected = 1'b0;
      for (addr = 'h00; addr < 'h800; addr += 1) begin
           expected = rand_wdat - addr;
           t_get_device_mem(addr, from_mem);
           if (from_mem !== expected) begin
              error_detected = 1'b1;
            end
       end

       if (error_detected)
         t_compare_memories('h00000000, 'h0, 2048, badcomp, goodcomp, 1);

       if (error_detected != 1'b1) begin
            $write("PASS \n");
       end else begin
            $write("FAIL  -- IRQ was OK But data integrity failed!!!\n");
       end

      #1us;
  endtask





// -------------------------------------------
//  TESTCASE:    t_test_8KiB_Byteswaps
//      Purpose:
//            tests byteswap behaviour
//      Method:
//          .....
// --------------------------------------------
   task automatic t_test_8KiB_Byteswaps(ref   int ntest,
                                        input int tests
                                       );
      uint32_t addr, val, expected, from_mem;
      bit      error_detected;
      uint32_t badcomp, goodcomp;

      // Ramdomise the write data for this testcase, that way a previous test can't cause a false +ve..
      logic [31:0] rand_wdat = $urandom();

      for (int i = 0; i < 4; i++) begin

          $sformat(testphase, "8KiB DMA read (byte swap = %0d) (rand = 0x%08x) : ",
                                                          i,     rand_wdat
                  );
          $write("Test %0d/%0d: %s", ntest++, tests, testphase);



         for (addr = 'h00; addr < 'h800; addr += 1) begin
           t_blat_device_mem(addr, rand_wdat - addr);
         end


        t_setup_dma(  .host_addr               ('h2000_0000)
                     ,.local_addr              ('h00)
                     ,.length                  ('h2000)
                     ,.more_in_chain           (0)
                     ,.w_nrd_sel               (0)
                     ,.next_chained_descr_addr ('hffeff_0123_4567_19a0)
                   );

         acc.write('h00, (i << 2) | 'h01); // start    BUT byteswaps are applied...

         @(posedge dma_irq);

         check_irq_status;


         for (addr = 'h00; addr < 'h800; addr += 1)
           begin
              expected = rand_wdat - addr;
              if (i == 1)
                expected = {<<8{expected}};
              else if (i == 2)
                expected = {<<16{expected}};
              else if (i == 3)
                expected = {<<16{{<<8{expected}}}};
              mem_check(4 * addr, expected);
           end

         clear_irq;

         repeat(4) @(posedge clk_125m);

         $write("PASS\n");
          rand_wdat = $urandom();

         #1us;
      end
   endtask




// -------------------------------------------
//  TESTCASE:    t_test_mixed_chaining
//      Purpose:
//            Sanity checking a longer chain, and DMA direction changes...
//      Method:
//          NOTE: SELF_CHECKING: YES
//          The purpose of this testcase is to detect wishbone R->W and W->R turnarround issues from
//          the dma_controller state machine
//          as of 20/01/2021 it is failing (expected) ...
// --------------------------------------------
   task automatic t_test_mixed_chaining(ref   int ntest,
                                        input int tests
                                       );

      uint32_t     addr,expected;
      uint32_t     badcomp, goodcomp;
      logic [31:0] rand_wdat = $urandom();

     // logic [31:0] wdat;
      bit          dma_done;
      integer      cycles;

      uint32_t op_len_bytes = 128;

      uint64_t  wdat;
      uint64_t  start_addr;

      testphase = " t_test_mixed_chaining: ";
      $write("Test %0d/%0d: %s", ntest++, tests, testphase);

      // initialise memory



      for (addr = 'h00; addr < op_len_bytes/4; addr += 1) begin
        t_blat_device_mem(addr, rand_wdat - addr);
       // wdat = 64'h0;
       // wdat[0+:32] = rand_wdat - addr;
       // i_gn4124.host_mem_write_noaddr_twiddles(64'h2000_0000 + addr, wdat);
      end



      if (dma_irq != 1'b0) begin
        $fatal(1, "dma irq should be 0");
      end


      // First DMA copies tha data into system memory at offset 0x1000
        t_setup_dma(  .host_addr               ('h2000_1000)
                     ,.local_addr              ('h00)
                     ,.length                  (op_len_bytes)
                     ,.more_in_chain           (1)
                     ,.w_nrd_sel               (0)
                     ,.next_chained_descr_addr ('h20020000)
                   );
      // SEcond DMA Wipes  teh region the data came from.
      t_add_chaining_descriptor(.start_address_in_host_memory(64'h2002_0000), .next_chained_descr_addr('h2002_0020), .length(op_len_bytes), .more_in_chain(1'b1), .w_nrd_sel(1'b1), .host_addr(64'h2000_0000), .local_addr('h0000));

      t_add_chaining_descriptor(.start_address_in_host_memory(64'h2002_0020), .next_chained_descr_addr('h2002_0040), .length(op_len_bytes/4), .more_in_chain(1'b1), .w_nrd_sel(1'b1), .host_addr(64'h2000_1000), .local_addr('h00));
      t_add_chaining_descriptor(.start_address_in_host_memory(64'h2002_0040), .next_chained_descr_addr('h2002_0060), .length(op_len_bytes/4), .more_in_chain(1'b1), .w_nrd_sel(1'b0), .host_addr(64'h2000_1000), .local_addr('h00));
      t_add_chaining_descriptor(.start_address_in_host_memory(64'h2002_0060), .next_chained_descr_addr('h2002_0080), .length(op_len_bytes/4), .more_in_chain(1'b1), .w_nrd_sel(1'b1), .host_addr(64'h2000_1000 +   op_len_bytes/4 ), .local_addr('h00 +   op_len_bytes/4));
      t_add_chaining_descriptor(.start_address_in_host_memory(64'h2002_0080), .next_chained_descr_addr('h2002_00a0), .length(op_len_bytes/4), .more_in_chain(1'b1), .w_nrd_sel(1'b0), .host_addr(64'h2000_1000 +   op_len_bytes/4 ), .local_addr('h00 +   op_len_bytes/4));
      t_add_chaining_descriptor(.start_address_in_host_memory(64'h2002_00a0), .next_chained_descr_addr('h2002_00c0), .length(op_len_bytes/4), .more_in_chain(1'b1), .w_nrd_sel(1'b1), .host_addr(64'h2000_1000 +   op_len_bytes/2 ), .local_addr('h00 +   op_len_bytes/2));
      t_add_chaining_descriptor(.start_address_in_host_memory(64'h2002_00c0), .next_chained_descr_addr('h2002_00e0), .length(op_len_bytes/4), .more_in_chain(1'b1), .w_nrd_sel(1'b0), .host_addr(64'h2000_1000 +   op_len_bytes/2 ), .local_addr('h00 +   op_len_bytes/2));
      t_add_chaining_descriptor(.start_address_in_host_memory(64'h2002_00e0), .next_chained_descr_addr('h2002_0100), .length(op_len_bytes/4), .more_in_chain(1'b1), .w_nrd_sel(1'b1), .host_addr(64'h2000_1000 + 3*op_len_bytes/4 ), .local_addr('h00 + 3*op_len_bytes/4));
      t_add_chaining_descriptor(.start_address_in_host_memory(64'h2002_0100), .next_chained_descr_addr('hdead_beed), .length(op_len_bytes/4), .more_in_chain(1'b0), .w_nrd_sel(1'b0), .host_addr(64'h2000_1000 + 3*op_len_bytes/4 ), .local_addr('h00 + 3*op_len_bytes/4));


      dma_done = 1'b0;
      fork
         begin
            t_start_dma();
            @(posedge dma_irq);
            dma_done = 1'b1;
            check_irq_status;
            clear_irq;
         end
         while (1'b0 == dma_done) begin

// Wait for a transfer to be done....     but to simulate something like DDR refresh timing backpressure lets wait a while...
// Trying to backport this, I hope the semantics are right.....
            @(posedge clk_125m);
            if (1'b1 == DUT.cmp_wrapped_gn4124.gen_with_dma.cmp_p2l_dma_master.dma_ctrl_start_p2l_i) begin
               @(posedge DUT.cmp_wrapped_gn4124.gen_with_dma.cmp_p2l_dma_master.dma_ctrl_done_o);
               @(posedge clk_125m);
               force DUT.cmp_wrapped_gn4124.dma_stall_i = 1'b1;
               @(posedge clk_125m);
               force DUT.cmp_wrapped_gn4124.dma_ack_i = 1'b0;
               repeat(125) @(posedge clk_125m);
               release DUT.cmp_wrapped_gn4124.dma_stall_i;
               @(posedge clk_125m);
               release DUT.cmp_wrapped_gn4124.dma_ack_i;
            end
         end
      join



      // allow wishbone to finish
     //#1us;
     // t_compare_memories('h00000000, 'h0, op_len_bytes/4, badcomp, goodcomp, 1);

      for (addr = 'h00; addr < op_len_bytes/4; addr += 1)
        begin
           expected = rand_wdat - addr;
           t_get_device_mem(addr, wdat);
           if (expected !== wdat) begin
              $write("\n  Exected 0x%08x, got 0x%08x, for addr 0x%04x \n", expected, wdat, addr);
              $fatal();
           end
        end

        // compare other mem regions as well!!

      repeat(2) @(posedge clk_125m);


      $write("PASS\n");
   endtask




// -------------------------------------------
//  TESTCASE:    t_test_32bit_addr_overflow_wr
//      Purpose:
//            Test a 32bit Overflow for the host address.
//      Method:
//            setup the 2 test infrastructure BAR's on different consecutive addrs,
//             (and lets be evil and make one on a 32bit bar and one a 64bit BAR...)
//              set up a small DMA that crosses this boundary and then check the data...
//      NOTES: TBD: this test passes, but the request corsses a 4KB boundary
//                   (actually it is a 32 -> 64 bit addr boundary).
//                 I epxect it should FAIL however the BFM is not flagging any errors.....   TBD
// --------------------------------------------
   task automatic t_test_32bit_addr_overflow_wr(ref   int ntest,
                                                input int tests
                                               );

      uint32_t     addr;
      uint32_t     expected;
      uint64_t     wdat;
      uint64_t     pcie_addr;
      uint64_t     rdat;
      automatic logic [31:0] rand_wdat = $urandom();

      testphase = "t_test_32bit_addr_overflow_wr: ";
      $write("Test %0d/%0d: %s", ntest++, tests, testphase);


      test_subphase = "config_bars";
      i_gn4124.t_configure_bar(0, 'h100000000, 'h20000000);
      i_gn4124.t_configure_bar(1, 'h0fffff000, 'h00001000);


      test_subphase = "init test mem";
      for (addr = 'h00; addr < 16; addr += 1) begin
         wdat = 64'h0;
         wdat[0+:32] = rand_wdat - addr;
         pcie_addr      = 64'hffff_fff0 + addr*4;
         i_gn4124.host_mem_write_noaddr_twiddles(pcie_addr, wdat);
      end

      test_subphase = "config_dma";
      t_setup_dma(  .host_addr               ('hffff_fff0)
                   ,.local_addr              ('h00)
                   ,.length                  ('h20)
                   ,.more_in_chain           (0)
                   ,.w_nrd_sel               (1)
                   ,.next_chained_descr_addr ('h20020000)
                 );

      test_subphase = "do_dma";
      t_start_dma();
      @(posedge dma_irq);
      check_irq_status;
      clear_irq;

      test_subphase = "Check_result";

     // t_compare_memories('hffff_fff0, 'h0, 8, addr, expected, 1'b1, 1'b1);


      // do data check
      for (addr = 'h00; addr < 8; addr++)
        begin
           expected = rand_wdat - addr;
           t_get_device_mem(addr, rdat);
           //$write("\n  Exected 0x%08x, got 0x%08x, for addr 0x%04x \n", expected, rdat, addr);
           if (expected !== rdat[31:0]) begin
              $write("\n ERROR: Exected 0x%08x, got 0x%08x, for addr 0x%04x \n", expected, rdat, addr);
              $fatal();
           end
        end


      test_subphase = "Revert BAR configs";
      i_gn4124.t_detault_bar_config();


      repeat(2) @(posedge clk_125m);
      $write("PASS\n");
   endtask




// -------------------------------------------
//  TESTCASE:    t_test_32bit_addr_overflow_rd
//      Purpose:
//            Test a 32bit Overflow for the host address.
//      Method:
//            setup the 2 test infrastructure BAR's on different consecutive addrs,
//             (and lets be evil and make one on a 32bit bar and one a 64bit BAR...)
//              set up a small DMA that crosses this boundary and then check the data...
// --------------------------------------------
   task automatic t_test_32bit_addr_overflow_rd(ref   int ntest,
                                                input int tests
                                               );

      uint32_t     addr;
      uint32_t     expected;
      uint64_t     wdat;
      uint64_t     pcie_addr;
      uint64_t     rdat;

      uint64_t    pcie_start_addr = 64'hffff_ffe8;
      automatic logic [31:0] rand_wdat = $urandom();

      testphase = "t_test_32bit_addr_overflow_rd: ";
      $write("Test %0d/%0d: %s", ntest++, tests, testphase);


      test_subphase = "config_bars";
      i_gn4124.t_configure_bar(0, 'h100000000, 'h20000000);
      i_gn4124.t_configure_bar(1, 'h0fffff000, 'h00001000);


      test_subphase = "init test mem";
      for (addr = 'h00; addr < 16; addr += 1) begin
         wdat = 64'h0;
         wdat[0+:32] = rand_wdat - addr;
         t_blat_device_mem(addr, wdat[31:0]);
      end

      test_subphase = "config_dma";
      t_setup_dma(  .host_addr               (pcie_start_addr)
                   ,.local_addr              ('h00)
                   ,.length                  ('h40)
                   ,.more_in_chain           (0)
                   ,.w_nrd_sel               (0)
                   ,.next_chained_descr_addr ('h20020000)
                 );

      test_subphase = "do_dma";
      t_start_dma();
      @(posedge dma_irq);
      check_irq_status;
      clear_irq;

      test_subphase = "Check_result";

   //   t_compare_memories(pcie_start_addr, 'h0, 16, addr, expected, 1'b1, 1'b1);


      // do data check
      for (addr = 'h00; addr < 16; addr++)
        begin
           expected = rand_wdat - addr;
           pcie_addr = pcie_start_addr + addr*4;
           i_gn4124.host_mem_read_noaddr_twiddles(pcie_addr, rdat);
          // $write("\n  Exected 0x%08x, got 0x%08x, for addr 0x%04x \n", expected, rdat, addr);
           if (expected !== rdat[31:0]) begin
              $write("\n ERROR: Exected 0x%08x, got 0x%08x, for addr 0x%04x \n", expected, rdat, addr);
              $fatal();
           end
        end


      test_subphase = "Revert BAR configs";
      i_gn4124.t_detault_bar_config();

      repeat(2) @(posedge clk_125m);
      $write("PASS\n");
   endtask


